"""numdisplay: Package for displaying numpy arrays in IRAF-compatible
                image display tool such as DS9 or XIMTOOL.

    This package provides several methods for controlling the display
    of the numpy array; namely,

        open(imtdev=None)::
            Open the default display device or the device specified
            in 'imtdev', such as 'inet:5137' or 'fifo:/dev/imt1o'.

        close()::
            Close the display device defined by 'imtdev'. This must
            be done before resetting the display buffer to a new size.

        display(pix, name=None, bufname=None, z1=None, z2=None, quiet=False, transform=None, scale=None, offset=None, frame=None)::
            Display the scaled array in display tool (ds9/ximtool/...).

        readcursor(sample=0)::
            Return a single cursor position from the image display.
            By default, this operation will wait for a keystroke before
            returning the cursor position. If 'sample' is set to 1,
            then it will NOT wait to read the cursor.
            This will return a string containing: x,y,frame and key.

        help()::
            print Version ID and this message.

    Notes
    -----
    Displaying a numpy array object involves:
        1.  Opening the connection to a usable display tool (such as DS9).

        2.  Setting the display parameters for the array, such as min and
            max array value to be used for min and max grey scale level, along
            with any offset, scale factor and/or transformation function to be
            applied to the array.
        3.  Applying any transformation to the input array.  This transformation
            could be a simple numpy ufunc or a user-defined function that
            returns a modified array.

        4.  Building the byte-scaled version of the transformed array and
            sending it to the display tool.  The image sent to the display
            device will be trimmed to fit the image buffer defined by the
            'imtdev' device from the 'imtoolrc' or the 'stdimage'
            variable under IRAF. If the image is smaller than the buffer,
            it will be centered in the display device.

            All pixel positions reported back will be relative to the
            full image size.

    Examples
    --------

    The user starts with a 1024x1024 array in the variable 'fdata'.
    This array has min pixel value of -157.04 and a max pixel value
    of 111292.02.  The display tool DS9 has already been started from
    the host level and awaits the array for display.  Displaying the
    array requires::

        >>> import numdisplay
        >>> numdisplay.display(fdata)

    If there is a problem connecting to the DS9 application, the connection
    can be manually started using::

        >>> numdisplay.open()

    To bring out the fainter features, an offset value of 158 can be added
    to the array to allow a 'log' scaling can be applied to the array values
    using::

        >>> numdisplay.display(fdata,transform=numpy.log,offset=158.0)

    To redisplay the image with default full-range scaling::

        >>> numdisplay.display(fdata)

    To redisplay using the IRAF display zscale algorithm, and with a contrast
    value steeper than the default value of 0.25::

        >>> numdisplay.display(fdata, zscale=True, contrast=0.5)


"""
from __future__ import division # confidence medium

from .version import *

import numpy as n
import math, string
import displaydev
import zscale as _zscale

try:
    import geotrans
except ImportError:
    geotrans = None


class NumDisplay(object):
    """ Class to manage the attributes and methods necessary for displaying
        the array in the image display tool.

        This class contains the methods:
            open(imtdev=None):

            close():

            set(z1=None,z2=None,scale=None,factor=None,frame=None):
            reset(names=None)

            display(pix, name=None, bufname=None):

            readcursor():

    """
    def __init__(self):
        self.frame = 1

        # Attributes used to scale image nearly arbitrarily
        # transform: name of Python function to operate on array
        #           default is to do nothing (apply self._noTransform)
        # scale : multiplicative factor to apply to input array
        # offset: additive factor to apply to input array

        self.transform = self._noTransform
        self.scale = None
        self.offset = None

        # default values for attributes used to determine pixel range values
        self.zscale = False
        self.stepline = 6
        self.contrast = 1    # Not implemented yet!
        self.nlines = 256    # Not implemented yet!

        # If zrange != 0, use user-specified min/max values
        self.zrange = 0  # 0 == False

        # Scale the image based on input pixel range...
        self.z1 = None
        self.z2 = None

        self.name = None
        self.view = displaydev._display
        self.handle = self.view.getHandle()

    def open(self,imtdev=None):
        """ Open a display device. """
        self.view.open(imtdev=imtdev)

    def close(self):
        """ Close the display device entry."""
        self.view.close()

    def set(self,frame=None,z1=None,z2=None,contrast=None,transform=None,scale=None,offset=None):

        """ Allows user to set multiple parameters at one time. """

        self.frame = frame
        self.zrange = 0  # 0 == False

        if contrast != None:
            self.contrast = contrast

        if z1 != None:
            self.z1 = z1
            self.zrange = 1  # 1 == True

        if z2 != None:
            self.z2 = z2
            self.zrange = 1  # 1 == True


        if transform:
            self.transform = transform

        if scale != None:
            self.scale = scale

        if offset is not None:
            self.offset = offset

    def reset(self, names=None):
        """Reset specific attributes, or all by default.

        Parameters
        ----------
        names : string or list of strings
            names of attributes to be reset, separated by commas
            or spaces; the default is to reset all attributes to None

        """

        if names:
            if isinstance(names, str):
                names = names.replace(",", " ")
                names = names.split()
            for name in names:
                if name == "transform":
                    self.transform = self._noTransform
                elif name == "zrange":
                    self.zrange = 0     # False
                else:
                    self.__setattr__(name, None)
            if "z1" in names or "z2" in names:
                self.zrange = 0         # False
        else:
            self.contrast = None
            self.z1 = None
            self.z2 = None
            self.transform = self._noTransform
            self.scale = None
            self.offset = None
            self.zrange = 0             # False

    def _noTransform(self, image):
        """ Applies NO transformation to image. Returns original.
            This will be the default operation when None is specified by user.
        """
        return image

    def _bscaleImage(self, image):
        """
        This function converts the input image into a byte-array
         with the z1/z2 values being mapped from 1 - 200.
         It also returns a status message to report on success or failure.

        """
        _pmin = 1.
        _pmax = 200.
        _ny,_nx = image.shape

        bimage = n.zeros((_ny,_nx),dtype=n.uint8)
        iz1 = self.z1
        iz2 = self.z2

        if iz2 == iz1:
            status = "Image scaled to all one pixel value!"
            return bimage
        else:
            scale =  (_pmax - _pmin) / (iz2 - iz1)

        # Now we can scale the pixels using a linear scale only (for now)
        # Scale the pixel values:  iz1 --> _pmin, iz2 --> _pmax
        bimage = (image - iz1) * scale + _pmin
        bimage = n.clip(bimage,_pmin,_pmax)
        bimage = n.array(bimage,dtype=n.uint8)

        status = 'Image scaled to Z1: '+repr(iz1)+' Z2: '+repr(iz2)+'...'
        return bimage


    def _fbclipImage(self,pix,fbwidth,fbheight):

        # Get the image parameters
        _ny,_nx = pix.shape

        if _nx > fbwidth or _ny > fbheight:

            # Compute the starting pixel of the image section to be displayed.
            _lx = (_nx // 2) - (fbwidth // 2)
            _ly = (_ny // 2) - (fbheight // 2)
            # We need to determine the region of the image to be put in frame
            _nx = min(_nx,fbwidth)
            _ny = min(_ny,fbheight)

        else:
            _lx = 0
            _ly = 0
        # Determine pixel range for image (sub)section
        # Insure it does not go beyond actual image array data bounds
        _xstart = max(_lx,0)
        _xend = max( (_lx + _nx),_nx)
        _ystart = max(_ly, 0)
        _yend = max( (_ly + _ny), _ny)

        # Return bytescaled, frame-buffer trimmed image
        if (_xstart == 0 and _xend == pix.shape[0]) and (_ystart == 0 and _yend == pix.shape[1]):
            return self._bscaleImage(pix)
        else:
            return self._bscaleImage(pix[_ystart:_yend,_xstart:_xend])

    def _transformImage(self, pix):
        """ Apply user-specified scaling to the input array. """

        if isinstance(pix,n.ndarray):

            if self.zrange:
                zpix = pix.copy()
                zpix = n.clip(pix,self.z1,self.z2)
            else:
                zpix = pix
        else:
            zpix = pix
        # Now, what kind of multiplicative scaling should be applied
        if self.scale:
            # Apply any additive offset to array
            if self.offset is not None:
                return self.transform( (zpix+self.offset)*self.scale)
            else:
                return self.transform( zpix*self.scale)
        else:
            if self.offset is not None:
                return self.transform (zpix + self.offset)
            else:
                return self.transform(zpix)

    def display(self, pix, name=None, bufname=None, z1=None, z2=None,
             transform=None, zscale=False, contrast=0.25, scale=None,
             offset=None, frame=None,quiet=False):

        """ Displays byte-scaled (UInt8) n to XIMTOOL device.
        This method uses the IIS protocol for displaying the data
        to the image display device, which requires the data to be
        byte-scaled.

        If input is not byte-scaled, it will perform scaling using
        set values/defaults.

        Parameters
        ----------
        name : str
            optional name to pass along for identifying array

        bufname : str
            name of buffer to use for displaying array (such as 'imt512').
            Other valid values include::

                'iraf': look for 'stdimage' and use that buffer or default to 'imt1024' [1024x1024 buffer]
                None  : ignore 'stdimage' and automatically select a buffer matched to the size of the image.

        z1,z2 : float
            minimum/maximum pixel value to display. Not specifying values will default
            to the full range values of the input array.

        transform : function
            Python function to apply to array (function)

        zscale : bool
            Specify whether or not to use an algorithm like that in the IRAF
            display task. If zscale=True, any z1 and z2 set in the call to display
            are ignored.  Using zscale=True invalidates any transform
            specified in the call.

        contrast : float (Default=0.25)
            Same as the *contrast* parameter in the IRAF *display* task.
            Only applies if zscale=True. Higher contrast values make z1 and
            z2 closer together, while lower values give a gentler (wider) range.

        scale : float/int
            multiplicative scale factor to apply to array. The value of this
            parameter remains persistent, so to reset it you must specify
            scale=1 in the display call.

        offset : float/int
            additive factor to apply to array before scaling. This value is
            persistent, so to reset it you have to set it to 0.

        frame : int
            image buffer frame number in which to display array

        quiet : bool (Default: False)
            if True, this parameter will turn off all status messages

        Notes
        ------
        The display parameters set here will ONLY apply to the display
        of the current array.

        """

        #Ensure that the input array 'pix' is a numpy array
        pix = n.array(pix)
        self.z1 = z1
        self.z2 = z2

        # If any of the display parameters are specified here, apply them
        #if z1 or z2 or transform or scale or offset or frame:
        # If zscale=True (like IRAF's display) selected, calculate z1 and z2 from
        # the data, and clear any transform specified in the call
        # Offset and scale arew applied to the data and z1,z2, so they have no effect
        # on the display
        if zscale:
            if transform != None:
                if not quiet:
                    print "transform disallowed when zscale=True"
                transform = None

            z1, z2 = _zscale.zscale(pix, contrast=contrast)

        self.set(frame=frame, z1=z1, z2=z2,
                transform=transform, scale=scale, offset=offset)

        # Initialize the display device
        if not self.view._display or self.view.checkDisplay() is False:
            self.open()
        _d = self.view._display
        self.handle = _d.getHandle()

        # If no user specified values are provided, interrogate the array itself
        # for the full range of pixel values
        if self.z1 == None:
            self.z1 = n.minimum.reduce(n.ravel(pix))
        if self.z2 == None:
            self.z2 = n.maximum.reduce(n.ravel(pix))

        # If the user has not selected a specific buffer for the display,
        # select and set the frame buffer size based on input image size.
        if bufname == 'iraf':
            useiraf = True
            bufname = None
        else:
            useiraf = False

        if bufname != None:
            _d.setFBconfig(None,bufname=bufname)
        else:
            _ny,_nx = pix.shape
            _d.selectFB(_nx,_ny,reset=1,useiraf=useiraf)

        # Initialize the specified frame buffer
        _d.setFrame(self.frame)
        _d.eraseFrame()

        # Apply user specified scaling to image, returns original
        # if none are specified.

        bpix = self._transformImage(pix)

        # Recompute the pixel range of (possibly) transformed array
        _z1 = self._transformImage(self.z1)
        _z2 = self._transformImage(self.z2)

        # If there was a problem in the transformation, then restore the original
        # array as the one to be displayed, even though it may not be ideal.
        if _z1 == _z2:
            if not quiet:
                print 'Error encountered during transformation. No transformation applied...'
            bpix = pix
            self.z1 = n.minimum.reduce(n.ravel(bpix))
            self.z2 = n.maximum.reduce(n.ravel(bpix))
            # Failsafe in case input image is flat:
            if self.z1 == self.z2:
                self.z1 -= 1.
                self.z2 += 1.
        else:
            # Reset z1/z2 values now so that image gets displayed with
            # correct range.  Also, when displaying transformed images
            # this allows the input pixel value to be displayed, rather
            # than the transformed pixel value.
            self.z1 = _z1
            self.z2 = _z2

        _wcsinfo = displaydev.ImageWCS(bpix,z1=self.z1,z2=self.z2,name=name)
        if not quiet:
            print 'Image displayed with Z1: ',self.z1,' Z2:',self.z2

        bpix = self._fbclipImage(bpix,_d.fbwidth,_d.fbheight)

        # Update the WCS to match the frame buffer being used.
        _d.syncWCS(_wcsinfo)

        # write out WCS to frame buffer, then erase buffer
        _d.writeWCS(_wcsinfo)

        # Now, send the trimmed image (section) to the display device
        _d.writeImage(bpix,_wcsinfo)
        #displaydev.close()

    def readcursor(self,sample=0):
        """ Return the cursor position from the image display. """
        return self.view.readCursor(sample=sample)

    def getHandle(self):
        return self.handle

    def checkDisplay(self):
        return self.view.checkDisplay()

# Help facility
def help():
    """ Print out doc string with syntax and example. """
    print 'numdisplay --- Version ',__version__
    print __doc__


view = NumDisplay()


# create aliases for PyDisplay methods
open = view.open
close = view.close

set = view.set
display = view.display
readcursor = view.readcursor
getHandle = view.getHandle
checkDisplay = view.checkDisplay

def sample() :
    '''stuff a sample image into the display

    use this to see that numdisplay is able to speak to the display program
    (ds9 or ximtool)
'''
    #
    numpy = n

    # an array of values 0..99
    a1 = numpy.arange(100)

    # an array going from 200..0 by 2
    a2 = numpy.arange(200,0,-2)

    # an empty array the size of the image I want
    b = numpy.zeros( 100 * 100 )

    # make the array square
    b.shape = ( 100, 100 )

    # copy a1 into the first 50 rows
    for x in range(0, 50) :
        b[x] = a1

    # copy a2 into the second 50 rows
    for x in range( 50, 100 ) :
        b[x] = a2

    # numdisplay.display(b)
    display(b)

    print "The first 50 rows are ascending brightness left  to right"
    print "The next  50 rows are ascending brightness right to left, but st"
    print "REMEMBER THAT 0,0 IS BOTTOM LEFT"



from __future__ import division # confidence high

import numpy as np
import os

def read_inc():
    """Read the iraf include file that defines the bit patterns for characters used
    for image display"""
    f = open('pixelfont.inc')
    lines = f.readlines()
    cdict = {}
    while len(lines) > 6:
        ic, bits, lines = nextchar(lines)
        cdict[ic] = bits
    return cdict
    
def nextchar(lines):
    """read next character data in file, and return numeric array 5x7 pixels"""
    if not lines[0].startswith('data'):
        raise ValueError
    else:
        clist = [lines[0][lines[0].find('/')+2:-3]]
        for i in range(1,7):
            clist.append(lines[i][:lines[i].find('B')].strip())
        ic = lines[6][lines[6].find('#')+1:].strip()
    carr
    return ic, clist, lines[8:]
        
def initichar_old():
    '''read the data file (old form) and generate the dict for numdisplay to use'''
    f = open('ichar_old.dat')
    fstr = f.read()
    cdict = eval(fstr)
    for key in cdict:
        clist = cdict[key]
        alist = []
        for item in clist:
            alist.append([int(c) for c in item])
        alist.reverse()
        cdict[key] = np.where(np.array(alist, dtype=np.bool_)) 
    return cdict 
    
def initichar():
    '''read the data file (old form) and generate the dict for numdisplay to use'''
    filepath = os.path.split(__file__)[0]
    f = open(filepath+'/ichar.dat')
    fstr = f.read()
    cdict = eval(fstr)
    return cdict
    
def expandchar(indices, size):
    '''block replicate a character. This implementation is inefficient
    but probably won't matter unless many large characters are used'''
    iy, ix = indices
    ox = np.zeros((size, size, len(indices[0]))) 
    oy = ox.copy()
    for i in range(size):
        for j in range(size):
            ox[i, j, :] = size*ix + i
            oy[i, j, :] = size*iy + j
    return (np.array(oy.flat, dtype=np.int32), np.array(ox.flat, dtype=np.int32))



from __future__ import division # confidence high

import math
import struct

import numpy as N
import stsci.numdisplay as numdisplay
import ichar

"""The public functions are the following.  For point, rectangle, circle
and polyline, arguments shown on separate lines are alternate ways to
specify the location and size of the figure to be drawn.

::

    point (x=x0, y=y0,
       center=(x0,y0))
    rectangle (left=x1, right=x2, lower=y1, upper=y2,
           center=(x0,y0), width=w, height=h)
    circle (x=x0, y=y0, radius=r,
        center=(x0,y0), radius=r)
    polyline (points=[(x1,y1), (x2,y2), (x3,y3), ...],
          vertices=[(x1,y1), (x2,y2), (x3,y3), ...])
    undo()
    set (color, radius)

Notes
-----
The *color* parameter is an optional argument to point, rectangle, circle, and polyline.

The allowed values for color are::
    C_BLACK, C_WHITE, C_RED, C_GREEN, C_BLUE, C_YELLOW, C_CYAN, C_MAGENTA,
    C_CORAL, C_MAROON, C_ORANGE, C_KHAKI, C_ORCHID, C_TURQUOISE, C_VIOLET, C_WHEAT
"""

C_BLACK     = 202
C_WHITE     = 203
C_RED       = 204
C_GREEN     = 205
C_BLUE      = 206
C_YELLOW    = 207
C_CYAN      = 208
C_MAGENTA   = 209
# the following all seem to display as magenta when using ds9
C_CORAL     = 210
C_MAROON    = 211
C_ORANGE    = 212
C_KHAKI     = 213
C_ORCHID    = 214
C_TURQUOISE = 215
C_VIOLET    = 216
C_WHEAT     = 217

# These two are used for saving the displayed values before drawing
# an overlay, to allow restoring the display (via undo).
global_save = []
global_byte_buf = N.zeros (1, dtype=N.uint8)

# These two are for convenience, so they can take default values rather
# than having to be specified for each function call.  The radius is
# only relevant for circles.  Either or both of these can be set via
# the set() function.
global_color = N.array ((C_CYAN,), dtype=N.uint8)
global_radius = 3

def _open_display(frame=1):
    """Open the device."""
    fd = numdisplay.getHandle()

    fd.setFrame(frame_num=frame)
    (tx, ty, fbwidth, fbheight) = fd.readInfo()
    return (fd, tx, ty, fbwidth, fbheight)

def close_display(frame=1):
    """Close the device."""
    fd = numdisplay.getHandle()
    fd.close()
    numdisplay.close()

def set (color=None, radius=None):
    """Specify the color or the radius.

    Parameters
    ----------
    color : int
        color code to use for graphic overlays; the allowed values (202..217) are::

          C_BLACK, C_WHITE, C_RED, C_GREEN, C_BLUE, C_YELLOW, C_CYAN, C_MAGENTA,
          C_CORAL, C_MAROON, C_ORANGE, C_KHAKI, C_ORCHID, C_TURQUOISE, C_VIOLET,
          C_WHEAT

    radius: int
        radius to use when drawing circles

    """

    global global_color, global_radius

    if color is not None:
        global_color = _checkColor (color)
    if radius is not None:
        if radius < 0:
            raise ValueError, "radius must be non-negative"
        global_radius = radius


def _transformPoint (x, y, tx, ty):
    """Convert image pixel coords to frame buffer coords (IIS convention).

    Parameters
    ----------
    x : float
        image X coordinate of point
    y : float
        image Y coordinate of point
    tx : float
        X offset of frame buffer in image (positive if image is larger)
    ty : float
        image coordinate corresponding to top line of frame buffer

    """

    x -= tx
    y = ty - y
    return (x, y)

def _checkColor (color=None):
    """Return a valid color.

    Parameters
    ----------
    color : int
        color code to use; if color=None, use default

    """

    global global_color

    if color is None:
        color = global_color
    elif color < C_BLACK or color > C_WHEAT:
        raise ValueError, "%d is not a valid color" % color
    else:
        color = N.array ((color,), dtype=N.uint8)
    return color

def _update_save (fd, x, y, list_of_points, last_overlay, undo=True):
    """Save info in local lists list_of_points and last_overlay.

    @param fd: for reading from image display
    @type fd: file handle
    @param x: X coordinate (IIS convention, not image coordinates)
    @type x: int
    @param y: Y coordinate (IIS convention, not image coordinates)
    @type y: int
    @param list_of_points: pixels that have been written to by the graphic
        overlay that called this function (updated by this function)
    @type list_of_points: list of (x,y) tuples
    @param last_overlay: pixel coordinates and current value of display
        (updated by this function)
    @type last_overlay: list of (x,y,value) tuples
    """

    global global_byte_buf
    if undo:
        if (x, y) not in list_of_points:
            value = fd.readData (x, y, global_byte_buf)
            value = struct.unpack ('B', value)
            list_of_points.append ((x, y))
            last_overlay.append ((x, y, value[0]))

def point (**kwargs):
    """Draw a point.

    Parameters
    ----------
    x : int
        image X coordinate of point
    y : int
        image Y coordinate of point
    center : tuple
        (x,y) coordinates of point
    color : int
        color code to use; if not specified, use default
    undo : bool
        if specified [default=True], keep track of overlays for undo()

    Examples
    --------
    Samples illustrating the syntax include::

        overlay.point (x=x0, y=y0)
        overlay.point (center=(x0,y0))
        overlay.point (x=x0, y=y0, color=overlay.C_<color>)

    """

    # These are used for saving what is currently displayed, for use by
    # the undo() function.
    global global_save, global_byte_buf
    last_overlay = []
    list_of_points = []

    allowed_arguments = ["x", "y", "center", "color", "frame", "undo"]
    x = None; y = None; center = None; color = None; frame = None; undo = True
    keys = kwargs.keys()
    if "center" in keys and ("x" in keys or "y" in keys):
        raise ValueError, \
            "Specify either 'center' or 'x' and 'y', but not both."
    for key in keys:
        if key in allowed_arguments:
            if key == "center":
                (x, y) = kwargs["center"]
            elif key == "x":
                x = kwargs["x"]
            elif key == "y":
                y = kwargs["y"]
            elif key == "color":
                color = kwargs["color"]
            elif key == "frame":
                frame = kwargs["frame"]
            elif key == "undo":
                undo = kwargs["undo"]
        else:
            raise ValueError, \
            "Invalid argument to 'point'; use 'x', 'y', 'center', 'color', 'frame' or 'undo'."
    if x is None or y is None:
        raise ValueError, "You must specify either 'x' and 'y' or 'center'."

    color = _checkColor (color)

    (fd, tx, ty, fbwidth, fbheight) = _open_display(frame=frame)

    (x, y) = _transformPoint (x, y, tx, ty)
    if x >= 0 and y >= 0 and x < fbwidth and y < fbheight:
        # save the value that is currently at (x,y)
        _update_save (fd, x, y, list_of_points, last_overlay,undo=undo)
        # write a new value at (x,y)
        fd.writeData (x, y, color)

    global_save.append (last_overlay)

    # The close() method needs to be called by the calling routine.
    #fd.close()

def marker (**kwargs):
    """Draw a character.

    Parameters
    ----------
    x : int
        image X coordinate of point
    y : int
        image Y coordinate of point
    mark : str
        character to be drawn
    size : int
        magnification to be used in drawing the character
    color : int
        color code to use; if not specified, use default
    undo : bool
        if specified [default=True], keep track of overlays for undo()

    Examples
    --------
    Samples illustrating the syntax include::

        overlay.marker (x=x0, y=y0, mark='+')
        overlay.marker (x=x0, y=y0, mark='+', size=2)
        overlay.marker (x=x0, y=y0, mark='+', color=overlay.C_<color>)
    """

    # These are used for saving what is currently displayed, for use by
    # the undo() function.
    global global_save, global_byte_buf
    last_overlay = []
    list_of_points = []

    allowed_arguments = ["x", "y", "mark", "color", "frame", "size", "undo"]
    x = None; y = None; center = None; color = None; frame = None; undo=True
    keys = kwargs.keys()

    for key in keys:
        if key in allowed_arguments:
            if key == "x":
                x = kwargs["x"]
            elif key == "y":
                y = kwargs["y"]
            elif key == "color":
                color = kwargs["color"]
            elif key == "mark":
                mark = kwargs["mark"]
            elif key == "size":
                txsize = kwargs["size"]
            elif key == "frame":
                frame = kwargs["frame"]
            elif key == "undo":
                undo = kwargs["undo"]
        else:
            raise ValueError, \
            "Invalid argument to 'point'; use 'x', 'y', 'mark', 'size', 'color', 'frame' or 'undo'."
    if x is None or y is None:
        raise ValueError, "You must specify 'x' and 'y'."

    color = _checkColor (color)

    (fd, tx, ty, fbwidth, fbheight) = _open_display(frame=frame)

    (x, y) = _transformPoint (x, y, tx, ty)

    idict = ichar.initichar()
    sprite = idict[mark]
    # Should have the '+' for center of image
    sprite = ichar.expandchar(sprite, txsize)
    ixsize = 5*txsize
    iysize = 7*txsize
    # Simple version: just use overlay with a thickness of 1
    points = sprite
    npts = len(points[0])
    for i in range(npts):
        iy = y - iysize//2 + points[0][i]
        ix = x - ixsize//2 + points[1][i]

        if ix >= 0 and iy >= 0 and ix < fbwidth and iy < fbheight:
            # save the value that is currently at (x,y)
            _update_save (fd, ix, iy, list_of_points, last_overlay,undo=undo)
            # write a new value at (x,y)
            fd.writeData (ix, iy, color)

    global_save.append (last_overlay)

    # The close() method needs to be called by the calling routine.
    #fd.close()

def rectangle (**kwargs):
    """Draw a rectangle.

    Parameters
    ----------
    left : int
        image X coordinate of left edge
    right : int
        image X coordinate of right edge
    lower : int
        image Y coordinate of lower edge
    upper : int
        image Y coordinate of upper edge
    center : tuple
        (x,y) coordinates of middle of rectangle
    width : int
        width of rectangle (X direction)
    height : int
        height of rectangle (Y direction)
    color : int
        color code to use; if not specified, use default
    undo : bool
        if specified [default=True], keep track of overlays for undo()

    Examples
    --------
    Samples illustrating the syntax include::

        overlay.rectangle (left=x1, right=x2, lower=y1, upper=y2)
        overlay.rectangle (center=(x0,y0), width=w, height=h)
        overlay.rectangle (left=x1, lower=y1, center=(x0,y0))
        overlay.rectangle (left=x1, lower=y1, width=w, height=h)
        overlay.rectangle (right=x2, upper=y2, width=w, height=h)
        overlay.rectangle (right=x2, upper=y2, center=(x0,y0))
        overlay.rectangle (left=x1, right=x2, lower=y1, upper=y2, color=overlay.C_<color>)

    """

    # These are used for saving what is currently displayed, for use by
    # the undo() function.
    global global_save, global_byte_buf
    last_overlay = []
    list_of_points = []

    allowed_arguments = ["left", "right", "lower", "upper",
                         "center", "width", "height", "color", "undo"]
    x1 = None; x2 = None; y1 = None; y2 = None
    center = None; width = None; height = None; color = None; undo = True

    error_message = "Invalid argument(s) to 'rectangle'; use one of:\n" \
"  left=x1, right=x2, lower=y1, upper=y2, or\n" \
"  center=(x0,y0), width=w, height=h, or\n" \
"  left=x1, lower=y1, width=w, height=h, or\n" \
"  right=x2, upper=y2, width=w, height=h\n" \
"  color and undo may also be specified."

    keys = kwargs.keys()
    if "center" in keys:
        if "width" not in keys or "height" not in keys:
            raise ValueError, error_message
    if "left" in keys:
        if "right" not in keys and "width" not in keys:
            raise ValueError, error_message
    if "lower" in keys:
        if "upper" not in keys and "height" not in keys:
            raise ValueError, error_message
    if "right" in keys:
        if "left" not in keys and "width" not in keys:
            raise ValueError, error_message
    if "upper" in keys:
        if "lower" not in keys and "height" not in keys:
            raise ValueError, error_message

    for key in keys:
        if key in allowed_arguments:
            if key == "center":
                center = kwargs["center"]
                if not isinstance (center, (list, tuple)):
                    raise ValueError, error_message
                (x0, y0) = center
            elif key == "left":
                x1 = kwargs["left"]
            elif key == "right":
                x2 = kwargs["right"]
            elif key == "lower":
                y1 = kwargs["lower"]
            elif key == "upper":
                y2 = kwargs["upper"]
            elif key == "width":
                width = kwargs["width"]
            elif key == "height":
                height = kwargs["height"]
            elif key == "color":
                color = kwargs["color"]
            elif key == "undo":
                undo = kwargs["undo"]
        else:
            raise ValueError, error_message

    if x1 is None:
        if center is not None and width is not None:
            x1 = int (round (x0 - width/2.))
        elif x2 is not None and width is not None:
            x1 = x2 - width
    if x2 is None:
        if center is not None and width is not None:
            x2 = int (round (x0 + width/2.))
        elif x1 is not None and width is not None:
            x2 = x1 + width
    if y1 is None:
        if center is not None and height is not None:
            y1 = int (round (y0 - height/2.))
        elif y2 is not None and height is not None:
            y1 = y2 - height
    if y2 is None:
        if center is not None and height is not None:
            y2 = int (round (y0 + height/2.))
        elif y1 is not None and height is not None:
            y2 = y1 + height

    if x1 is None or x2 is None or y1 is None or y2 is None:
        raise ValueError, error_message

    color = _checkColor (color)

    (fd, tx, ty, fbwidth, fbheight) = _open_display()

    (x1, y1) = _transformPoint (x1, y1, tx, ty)
    (x2, y2) = _transformPoint (x2, y2, tx, ty)

    if x2 < x1:
        (x1, x2) = (x2, x1)
    if y2 < y1:
        (y1, y2) = (y2, y1)

    imin = max (0, x1)
    imax = min (x2+1, fbwidth)
    if y1 >= 0 and y1 < fbheight:
        for i in range (imin, imax):
            _update_save (fd, i, y1, list_of_points, last_overlay, undo=undo)
            fd.writeData (i, y1, color)
    if y2 >= 0 and y2 < fbheight:
        for i in range (imin, imax):
            _update_save (fd, i, y2, list_of_points, last_overlay, undo=undo)
            fd.writeData (i, y2, color)

    jmin = max (0, y1)
    jmax = min (y2+1, fbheight)
    if x1 >= 0 and x1 < fbwidth:
        for j in range (jmin, jmax):
            _update_save (fd, x1, j, list_of_points, last_overlay, undo=undo)
            fd.writeData (x1, j, color)
    if x2 >= 0 and x2 < fbwidth:
        for j in range (jmin, jmax):
            _update_save (fd, x2, j, list_of_points, last_overlay, undo=undo)
            fd.writeData (x2, j, color)

    global_save.append (last_overlay)

    # The close() method needs to be called by the calling routine.
    #fd.close()

def circle (**kwargs):
    """Draw a circle.

    Parameters
    ----------
    x : int
        image X coordinate of center
    y : int
        image Y coordinate of center
    center : tuple
        (x,y) coordinates of center
    radius : int
        radius of circle
    color : int
        color code to use; if not specified, use default
    undo : bool
        if specified [default=True], keep track of overlays for undo()

    Examples
    --------
    Samples illustrating the syntax include::

        overlay.circle (x=x0, y=y0, radius=r)
        overlay.circle (center=(x0,y0), radius=r)
        overlay.circle (x=x0, y=y0, radius=r, color=overlay.C_<color>)

    """

    # These are used for saving what is currently displayed, for use by
    # the undo() function.
    global global_save, global_byte_buf
    last_overlay = []
    list_of_points = []

    allowed_arguments = ["x", "y", "center", "radius", "color", "frame", "undo"]
    x0 = None; y0 = None; center = None;
    radius = global_radius; color = None; frame = None; undo = True

    error_message = "Invalid argument(s) to 'circle'; use either:\n" \
"  x=x0, y=x0, radius=r, or\n" \
"  center=(x0,y0), radius=r\n" \
"  color, frame and undo may also be specified."

    keys = kwargs.keys()
    if "center" in keys and ("x" in keys or "y" in keys):
        raise ValueError, error_message
    for key in keys:
        if key in allowed_arguments:
            if key == "center":
                center = kwargs["center"]
                if not isinstance (center, (list, tuple)):
                    raise ValueError, error_message
                (x0, y0) = center
            elif key == "x":
                x0 = kwargs["x"]
            elif key == "y":
                y0 = kwargs["y"]
            elif key == "radius":
                radius = kwargs["radius"]
            elif key == "color":
                color = kwargs["color"]
            elif key == "frame":
                frame = kwargs["frame"]
            elif key == "undo":
                undo = kwargs["undo"]
        else:
            raise ValueError, error_message
    if x0 is None or y0 is None or radius is None:
        raise ValueError, error_message

    color = _checkColor (color)

    (fd, tx, ty, fbwidth, fbheight) = _open_display(frame=frame)

    (x0, y0) = _transformPoint (x0, y0, tx, ty)
    quarter = int (math.ceil (radius * math.sqrt (0.5)))
    r2 = radius**2

    for dy in range (-quarter, quarter+1):
        dx = math.sqrt (r2 - dy**2)
        j = int (round (dy + y0))
        i = int (round (x0 - dx))           # left arc
        if i >= 0 and j >= 0 and i < fbwidth and j < fbheight:
            _update_save (fd, i, j, list_of_points, last_overlay, undo = undo)
            fd.writeData (i, j, color)
        i = int (round (x0 + dx))           # right arc
        if i >= 0 and j >= 0 and i < fbwidth and j < fbheight:
            _update_save (fd, i, j, list_of_points, last_overlay, undo = undo)
            fd.writeData (i, j, color)

    for dx in range (-quarter, quarter+1):
        dy = math.sqrt (r2 - dx**2)
        i = int (round (dx + x0))
        j = int (round (y0 - dy))           # bottom arc
        if i >= 0 and j >= 0 and i < fbwidth and j < fbheight:
            _update_save (fd, i, j, list_of_points, last_overlay, undo=undo)
            fd.writeData (i, j, color)
        j = int (round (y0 + dy))           # top arc
        if i >= 0 and j >= 0 and i < fbwidth and j < fbheight:
            _update_save (fd, i, j, list_of_points, last_overlay, undo=undo)
            fd.writeData (i, j, color)

    global_save.append (last_overlay)

    # The close() method needs to be called by the calling routine.
    #fd.close()

def polyline (**kwargs):
    """Draw a series of connected line segments.

    Parameters
    ----------
    points : list of tuples
        (x,y) points to connect with line segments
    vertices : list of tuples
        (x,y) points to connect with line segments
    color : int
        color code to use; if not specified, use default
    undo : bool
        if specified [default=True], keep track of overlays for undo()

    Examples
    --------
    Samples illustrating the syntax include::

        overlay.polyline (points=[(x1,y1), (x2,y2), (x3,y3)])
        overlay.polyline (vertices=[(x1,y1), (x2,y2), (x3,y3)])
        overlay.polyline (points=[(x1,y1), (x2,y2), (x3,y3)],
                          color=overlay.C_<color>)

    """

    # These are used for saving what is currently displayed, for use by
    # the undo() function.
    global global_save, global_byte_buf
    last_overlay = []
    list_of_points = []

    allowed_arguments = ["points", "vertices", "color", "frame", "undo"]
    points = None; vertices = None; color = None; frame = None; undo=True

    error_message = "Invalid argument(s) to 'polyline'; use either:\n" \
"  points=[(x1,y1), (x2,y2), (x3,y3), <...>] or\n" \
"  vertices=[(x1,y1), (x2,y2), (x3,y3), <...>]\n" \
"  color, frame, or undo may also be specified."

    keys = kwargs.keys()
    if "points" not in keys and "vertices" not in keys:
        raise ValueError, error_message
    for key in keys:
        if key in allowed_arguments:
            if key == "points":
                points = kwargs["points"]
                if not isinstance (points, (list, tuple)):
                    raise ValueError, error_message
            elif key == "vertices":
                vertices = kwargs["vertices"]
                if not isinstance (vertices, (list, tuple)):
                    raise ValueError, error_message
            elif key == "color":
                color = kwargs["color"]
            elif key == "frame":
                frame = kwargs["frame"]
            elif key == "undo":
                undo = kwargs["undo"]
        else:
            raise ValueError, error_message

    if points is not None and vertices is not None:
        raise ValueError, error_message
    if vertices is not None:
        keyword = "vertices"
        points = vertices
    else:
        keyword = "points"

    color = _checkColor (color)

    (fd, tx, ty, fbwidth, fbheight) = _open_display(frame=frame)

    expected_a_tuple = \
    "Each point in %s for polyline must be a two-element list or tuple,\n" \
    "giving the X and Y image pixel coordinates of a vertex."

    first = True
    for point in points:
        if not isinstance (point, (list, tuple)):
            raise ValueError, expected_a_tuple
        (x, y) = point
        (x, y) = _transformPoint (x, y, tx, ty)
        if first:
            (xlast, ylast) = (x, y)
            first = False
            continue
        if x == xlast and y == ylast:
            continue
        dx = x - xlast
        dy = y - ylast
        (x1, x2, y1, y2) = (xlast, x, ylast, y)
        if abs (dy) <= abs (dx):
            if x >= xlast:
                step = 1
            else:
                step = -1
            slope = float(dy) / float(dx)
            if step > 0:
                imin = max (0, x1)
                imax = min (x2+1, fbwidth)
            else:
                imin = min (x1, fbwidth-1)
                imax = max (x2-1, -1)
            for i in range (imin, imax, step):
                j = slope * (i - x1) + y1
                j = int (round (j))
                if j >= 0 and j < fbheight:
                    _update_save (fd, i, j, list_of_points, last_overlay, undo=undo)
                    fd.writeData (i, j, color)
        else:
            if y >= ylast:
                step = 1
            else:
                step = -1
            slope = float(dx) / float(dy)
            if step > 0:
                jmin = max (0, y1)
                jmax = min (y2+1, fbheight)
            else:
                jmin = min (y1, fbheight-1)
                jmax = max (y2-1, -1)
            for j in range (jmin, jmax, step):
                i = slope * (j - y1) + x1
                i = int (round (i))
                if i >= 0 and i < fbwidth:
                    _update_save (fd, i, j, list_of_points, last_overlay, undo=undo)
                    fd.writeData (i, j, color)
        xlast = x
        ylast = y

    global_save.append (last_overlay)

    # The close() method needs to be called by the calling routine.
    #fd.close()

def undo():
    """Restore the values before the last overlay was written."""

    global global_save, global_byte_buf

    if len (global_save) == 0:
        return

    (fd, tx, ty, fbwidth, fbheight) = _open_display()

    last_overlay = global_save.pop()
    for (x, y, value) in last_overlay:
        global_byte_buf[0] = value
        fd.writeData (x, y, global_byte_buf)

    # The close() method needs to be called by the calling routine.
    #fd.close()



"""displaydev.py: Interact with IRAF-compatible image display

Modeled after the NOAO Client Display Library (CDL)

Public functions:

readCursor(sample=0)
        Read image cursor position

open(imtdev=None)
        Open a connection to the display server.  This is called
        automatically by readCursor if the display has not already been
        opened, so it is not generally necessary for users to call it.

        See the open doc string for info on the imtdev argument, which
        allows various forms of socket and network connections.

close()
        Close the active display server.  Called automatically on exit.

Various classes are defined for the different connections (ImageDisplay,
ImageDisplayProxy, UnixImageDisplay, InetImageDisplay, FifoImageDisplay).
They should generally be created using the _open factory function.
This could be used to maintain references to multiple display servers.

Ultimately more functionality may be added to make this a complete
replacement for CDL.

$Id: displaydev.py 19347 2012-09-12 20:36:42Z hack $

"""
from __future__ import division # confidence medium

import os, socket, struct

import numpy as n
import string
import imconfig

try:
    # Only try to import this on Unix-compatible systems,
    # since Win platforms does not have the 'fcntl' library.
    import fcntl

    # FCNTL is deprecated in Python 2.2
    if hasattr(fcntl,'F_SETFL'):
        FCNTL = fcntl
    else:
        import FCNTL
except ImportError, error:
    fcntl = None
    # set a default that will be compatible with Win platform
    os.O_NDELAY = 0

try:
    from stsci.tools import fileutil
except:
    fileutil=None

try:
    SOCKTYPE = socket.AF_UNIX
except AttributeError, error:
    SOCKTYPE = socket.AF_INET

SZ_BLOCK = 16384

_default_imtdev = ("unix:/tmp/.IMT%d", "fifo:/dev/imt1i:/dev/imt1o","inet:5137")
_default_fbconfig = 3

class ImageWCS(object):
    _W_UNITARY = 0
    _W_LINEAR = 1
    _W_LOG = 2
    _W_USER = 3

    def __init__(self,pix,name=None,title=None,z1=None,z2=None):
        # pix must be a numpy array...
        self.a = 1.0
        self.b = self.c = 0.
        self.d = -1.0
        _shape = pix.shape
        # Start assuming full image can fit in frame buffer
        self.tx = int(_shape[1] / 2. + 0.5)
        self.ty = int(_shape[0] / 2. + 0.5)
        self.dtx = _shape[1] / 2.
        self.dty = _shape[0] / 2.

        # Determine full range of pixel values for image
        if not z1:
            self.z1 = n.minimum.reduce(n.ravel(pix))
        else:
            self.z1 = z1

        if not z2:
            self.z2 = n.maximum.reduce(n.ravel(pix))
        else:
            self.z2 = z2

        self.zt = self._W_LINEAR

        if not name:
            self.name = 'Image'
        else:
            self.name = name
        self.title = title

        self.ny,self.nx = pix.shape
        self.full_ny, self.full_nx = pix.shape

    def update(self,wcsstr):
        # This routine will accept output from readWCS and
        # update the WCS attributes with the values
        if wcsstr != None:
            _wcs = wcsstr.split()
            self.name = _wcs[0]
            self.a = float(_wcs[1])
            self.b = float(_wcs[2])
            self.c = float(_wcs[3])
            self.d = float(_wcs[4])
            self.tx = float(_wcs[5])
            self.ty = float(_wcs[6])
            self.z1 = float(_wcs[7])
            self.z2 = float(_wcs[8])
            self.zt = int(_wcs[9])


    def __str__(self):
        # This method can be used for obtaining the string
        # necessary for setting the WCS of the frame buffer
        # Start by insuring that the image name if not set is
        # passed as the string 'None' at least

        _name = self.name
        if self.title != None:
            _name = _name+'-'+repr(self.title)
        _name =_name+'\n'

        _str = _name+str(self.a)+' '+str(self.b)+' '+str(self.c)+' '+str(self.d)
        _str = _str+' '+str(self.tx)+' '+str(self.ty)
        _str = _str+' '+str(self.z1)+' '+str(self.z2)+' '+str(self.zt)+'\n'

        return(_str)


def _open(imtdev=None):

    """Open connection to the image display server

    This is a factory function that returns an instance of the ImageDisplay
    class for the specified imtdev.  The default connection if no imtdev is
    specified is given in the environment variable IMTDEV (if defined) or
    is "unix:/tmp/.IMT%d".  Failing that, a connection is attempted on the
    /dev/imt1[io] named fifo pipes.

    The syntax for the imtdev argument is <domain>:<address> where <domain>
    is one of "inet" (internet tcp/ip socket), "unix" (unix domain socket)
    or "fifo" (named pipe).  The form of the address depends upon the
    domain, as illustrated in the examples below.
    ::

     inet:5137                  Server connection to port 5137 on the local
                                host.  For a client, a connection to the
                                given port on the local host.

     inet:5137:foo.bar.edu      Client connection to port 5137 on internet
                                host foo.bar.edu.  The dotted form of address
                                may also be used.

     unix:/tmp/.IMT212          Unix domain socket with the given pathname
                                IPC method, local host only.

     fifo:/dev/imt1i:/dev/imt1o FIFO or named pipe with the given pathname.
                                IPC method, local host only.  Two pathnames
                                are required, one for input and one for
                                output, since FIFOs are not bidirectional.
                                For a client the first fifo listed will be
                                the client's input fifo; for a server the
                                first fifo will be the server's output fifo.
                                This allows the same address to be used for
                                both the client and the server, as for the
                                other domains.

    The address field may contain one or more "%d" fields.  If present, the
    user's UID will be substituted (e.g. "unix:/tmp/.IMT%d").
    """

    if not imtdev:
        # try defaults
        defaults = list(_default_imtdev)
        if os.environ.has_key('IMTDEV'):
            defaults.insert(0,os.environ['IMTDEV'])
        for imtdev in defaults:
            try:
                return _open(imtdev)
            except:
                pass
        raise IOError("Cannot attach to display program. Verify that one is running...")
    # substitute user id in name (multiple times) if necessary
    nd = len(imtdev.split("%d"))
    if nd > 1:
        dev = imtdev % ((abs(os.getpid()),)*(nd-1))
    else:
        dev = imtdev
    fields = dev.split(":")
    domain = fields[0]
    if domain == "unix" and len(fields) == 2:
        return UnixImageDisplay(fields[1])
    elif domain == "fifo" and len(fields) == 3:
        return FifoImageDisplay(fields[1],fields[2])
    elif domain == "inet" and (2 <= len(fields) <= 3):
        try:
            port = int(fields[1])
            if len(fields) == 3:
                hostname = fields[2]
            else:
                hostname = None
            return InetImageDisplay(port, hostname)
        except ValueError:
            pass
    raise ValueError("Illegal image device specification `%s'"
                                    % imtdev)


class ImageDisplay(object):

    """Interface to IRAF-compatible image display"""

    # constants for IIS Protocol header packets
    _IIS_READ =   0100000
    _IIS_WRITE =  0400000
    _COMMAND =    0100000
    _PACKED =     0040000
    _IMC_SAMPLE = 0040000

    _MEMORY = 01
    _LUT = 02
    _FEEDBACK = 05
    _IMCURSOR = 020
    _WCS = 021

    _SZ_IMCURVAL = 160
    _SZ_WCSBUF = 320

    def __init__(self):
        # Flag indicating that readCursor request is active.
        # This is used to handle interruption of readCursor before
        # read is complete.  Without this kluge, ^C interrupts
        # leave image display in a bad state.
        self._inCursorMode = 0

        # Add hooks here for managing frame configuration
        self.fbdict = imconfig.loadImtoolrc()

        self.frame = 1

        self.fbname = None

        _fbconfig = self.getDefaultFBConfig()

        self.fbconfig = _fbconfig
        self.fbwidth = self.fbdict[self.fbconfig]['width']
        self.fbheight = self.fbdict[self.fbconfig]['height']


    def getDefaultFBConfig(self):
        try:
            # Try to use the IRAF 'stdimage' value as the default
            # fbconfig number, if it exists
            if fileutil is not None:
                self.fbname = fileutil.envget('stdimage')
            else:
                if 'stdimage' in os.environ:
                    self.fbname = os.environ['stdimage']

            if self.fbname is not None:
                # Search through all IMTOOLRC entries to find a match
                _fbconfig = self.getConfigno(self.fbname)
            else:
                _fbconfig = _default_fbconfig
        except:
            _fbconfig = _default_fbconfig

        return _fbconfig

    def readCursor(self,sample=0):

        """Read image cursor value for this image display

        Return immediately if sample is true, or wait for keystroke
        if sample is false (default).  Returns a string with
        x, y, frame, and key.
        """

        if not self._inCursorMode:
            opcode = self._IIS_READ
            if sample:
                opcode |= self._IMC_SAMPLE
            self._writeHeader(opcode, self._IMCURSOR, 0, 0, 0, 0, 0)
            self._inCursorMode = 1
        s = self._read(self._SZ_IMCURVAL)
        self._inCursorMode = 0
        # only part up to newline is real data
        return s.split("\n")[0]

    def getConfigno(self,stdname):

        """ Determine which config number matches
        specified frame buffer name.
        """
        _fbconfig = None
        # Search through all IMTOOLRC entries to find a match
        for fb in self.fbdict.keys():
            if string.find(self.fbdict[fb]['name'],string.strip(stdname)) > 0:
                _fbconfig = int(fb)
                break
        if not _fbconfig:
            # If no matching configuration found,
            # default to 'imt1024'
            _fbconfig = _default_fbconfig

        return _fbconfig

    def selectFB(self,nx,ny,reset=None,useiraf=True):
        """ Select the frame buffer that best matches the input image size."""
        newfb = None
        _tmin = 100000

        #if iraffunctions:
        if self.fbname is not None and useiraf:
            # Use the STDIMAGE variable defined in the IRAF session...
            newfb = self.getDefaultFBConfig()
        else:
            # Otherwise, fall back to finding the buffer with the
            # size closest to the image's size.
            for fb in self.fbdict.keys():
                _fbw = self.fbdict[fb]['width']
                _fbh = self.fbdict[fb]['height']
                if nx == _fbw and ny == _fbh:
                    # Found an exact match.
                    newfb = fb
                    break
                elif _fbw > nx and _fbh > ny:
                    # No exact match, so look for match with smallest padding.
                    _edges = _fbw - nx + _fbh - ny
                    if _edges < _tmin:
                        _tmin = _edges
                        newfb = fb

        # If no new frame buffer was found that matched better than default...
        if not newfb:
            # use the default (probably 'imt512')
            newfb = self.fbconfig

        # If reset was specified, automatically set new config to this value.
        if reset:
            self.setFBconfig(newfb)

        # At the very least, return the config number found.
        return newfb


    def setFBconfig(self,fbnum,bufname=None):

        """ Set the frame buffer values for the given frame buffer name. """

        if bufname:
            self.fbconfig = self.getConfigno(bufname)
        else:
            self.fbconfig = fbnum

        self.fbwidth = self.fbdict[self.fbconfig]['width']
        self.fbheight = self.fbdict[self.fbconfig]['height']

    def writeData(self,x,y,pix):

        """ Writes out image data to x,y position in active frame. """

        opcode = self._IIS_WRITE | self._PACKED
        frame = 1 << (self.frame-1)
        nbytes = pix.size * pix.itemsize
        self._writeHeader(opcode,self._MEMORY, -nbytes, x, y, frame, 0)

        status = self._write(pix.tostring())
        return status

    def readData(self,x,y,pix):

        """ Reads data from x,y position in active frame."""

        opcode = self._IIS_READ | self._PACKED
        nbytes = pix.size * pix.itemsize
        frame = 1 << (self.frame-1)
        self._writeHeader(opcode,self._MEMORY, -nbytes, x, y, frame, 0)

        # Get the pixels now
        return self._read(nbytes)

    def setCursor(self,x,y,wcs):

        """ Moves cursor to specified position in frame. """

        self._writeHeader(self._IIS_WRITE, self._IMCURSOR,0,x,y,wcs,0)

    def setFrame(self,frame_num=1):

        """ Sets the active frame in frame buffer to specified value."""

        code = self._LUT | self._COMMAND
        self._writeHeader(self._IIS_WRITE, code, -1, 0, 0, 0, 0)

        # Update with user specified frame number
        if frame_num:
            self.frame = frame_num

        # Convert to bit-shifted value for IIS stream
        frame = 1 << (self.frame - 1)
        # Write out 2-byte value for frame number
        self._write(struct.pack('H',frame))

    def eraseFrame(self):

        """ Sends commands to erase active frame."""

        opcode = self._IIS_WRITE + self.fbconfig-1

        frame = 1 << (self.frame-1)
        self._writeHeader(opcode, self._FEEDBACK, 0,0,0,frame,0)

    def writeWCS(self,wcsinfo):

        """ Writes out WCS information for frame to display device."""

        _str = string.rstrip(str(wcsinfo))
        nbytes = len(_str)
        opcode = self._IIS_WRITE | self._PACKED
        frame = 1 << (self.frame-1)
        fbconfig = self.fbconfig - 1

        self._writeHeader(opcode,self._WCS, -nbytes, 0,0, frame, fbconfig)

        status = self._write(_str)

    def readWCS(self,wcsinfo):

        """ Reads WCS information from active frame of display device."""

        frame = 1 << (self.frame-1)

        self._writeHeader(self._IIS_READ, self._WCS, 0,0,0,frame,0)

        wcsinfo.update(self._read(self._SZ_WCSBUF))
        return wcsinfo

    def readInfo(self):
        """Read tx and ty from active frame of display device."""

        frame = 1 << (self.frame-1)

        self._writeHeader(self._IIS_READ, self._WCS, 0,0,0,frame,0)

        wcsstr = self._read(self._SZ_WCSBUF)
        _wcs = wcsstr.split()
        tx = int(round(float(_wcs[5])))
        ty = int(round(float(_wcs[6])))
        # print "debug: ", wcsstr

        return (tx, ty, self.fbwidth, self.fbheight)

    def syncWCS(self,wcsinfo):

        """ Update WCS to match frame buffer being used. """

        # Update WCS information with offsets into frame buffer for image
        wcsinfo.tx = int(((wcsinfo.full_nx + 1.0) / 2.) - ((self.fbwidth) / 2.) + 0.5)
        wcsinfo.ty = int((self.fbheight) + ((wcsinfo.full_ny / 2.) - (self.fbheight / 2.)) + 0.5)

        wcsinfo.nx = min(wcsinfo.full_nx, self.fbwidth)
        wcsinfo.ny = min(wcsinfo.full_ny, self.fbheight)

        # Keep track of the origin of the displayed, trimmed image
        # which fits in the buffer.
        wcsinfo.dtx = int((wcsinfo.nx / 2.) - ((self.fbwidth) / 2.) + 0.5)
        wcsinfo.dty = int((self.fbheight) + ((wcsinfo.ny / 2.) - (self.fbheight / 2.)) + 0.5)

    def writeImage(self,pix,wcsinfo):

        """ Write out image to display device in 32Kb sections."""

        _fbnum = self.fbconfig
        _fbw = self.fbdict[_fbnum]['width']
        _fbh = self.fbdict[_fbnum]['height']
        _nx,_ny = wcsinfo.nx,wcsinfo.ny
        _ty = wcsinfo.dty
        _tx = wcsinfo.dtx

        _nnx = min(_nx,_fbw)
        _nny = min(_ny,_fbh)

        # compute the range in output pixels the input image would cover
        # input image could be smaller than buffer size/output image size.
        _lx = (_fbw // 2) - (_nnx // 2)

        _lper_block = SZ_BLOCK // _fbw
        if _lper_block > 1: _lper_block = 1
        _nblocks = _nny // _lper_block

        # Flip image array so that (0,0) pixel is in upper left
        _fpix = pix[::-1,:]

        # Now, for each block, write out the image section
        if _lper_block == 1:
            # send each line of image to display
            for block in xrange(int(_nblocks)):
                _ydisp = _fbh - (_ty - block)
                self.writeData(_lx,_ydisp,_fpix[block,:])
        else:
            # display each line segment separately
            for block in xrange(int(_nblocks)):
                _y0 = block * _lper_block
                _ydisp = _fbh - (_ty - _y0)
                _xper_block = (_nx // (_nx * _lper_block))
                for xblock in xrange(int(_xper_block)):
                    _x0 = xblock * _xper_block
                    _xend = xblock + 1 * _xper_block
                    if _xend > _nx: _xend = _nx
                    self.writeData(_lx,_ydisp,_fpix[_y0,_x0:_xend])

        #Now pick up last impartial block
        _yend = int(_nblocks) * _lper_block
        if _yend < _ny:
            #self.writeData(_lx,_yend+_ty,_fpix[_yend:0,:])
            self.writeData(_lx,_yend+_ty,_fpix[:_yend,:])


    def _writeHeader(self,tid,subunit,thingct,x,y,z,t):

        """Write request to image display"""

        a = n.array([tid,thingct,subunit,0,x,y,z,t],dtype=n.uint16)
        # Compute the checksum
        sum = n.add.reduce(a,dtype=n.uint16)
        sum = 0xffff - (sum & 0xffff)
        a[3] = sum
        self._write(a.tostring())

    def close(self, os_close=os.close):

        """Close image display connection"""

        try:
            os_close(self._fdin)
        except (OSError, AttributeError):
            pass
        try:
            os_close(self._fdout)
        except (OSError, AttributeError):
            pass

    def getHandle(self):
        return self

    def _read(self, n):
        """Read n bytes from image display and return as string

        Raises IOError on failure.  If a Tkinter widget exists, runs
        a Tk mainloop while waiting for data so that the Tk widgets
        remain responsive.
        """
        try:
            return os.read(self._fdin, n)
        except (EOFError, IOError):
            raise IOError("Error reading from image display")

    def _write(self, s):
        """Write string s to image display

        Raises IOError on failure
        """
        try:
            n = len(s)
            while n>0:
                nwritten = self._socket.send(s[-n:])
                n -= nwritten
                if nwritten <= 0:
                    raise IOError("Error writing to image display")
        except OSError:
            raise IOError("Error writing to image display")


class FifoImageDisplay(ImageDisplay):

    """FIFO version of image display"""

    def __init__(self, infile, outfile):
        ImageDisplay.__init__(self)
        try:
            self._fdin = os.open(infile, os.O_RDONLY | os.O_NDELAY)
            if fcntl:
                fcntl.fcntl(self._fdin, FCNTL.F_SETFL, os.O_RDONLY)

            self._fdout = os.open(outfile, os.O_WRONLY | os.O_NDELAY)
            if fcntl:
                fcntl.fcntl(self._fdout, FCNTL.F_SETFL, os.O_WRONLY)
        except OSError, error:
            raise IOError("Cannot open image display (%s)" % (error,))

    def _write(self, s):
        """Write string s to image display

        Raises IOError on failure
        """
        try:
            n = len(s)
            while n>0:
                nwritten = os.write(self._fdout, s[-n:])
                n -= nwritten
                if nwritten <= 0:
                    raise IOError("Error writing to image display")
        except OSError:
            raise IOError("Error writing to image display")

    def __del__(self):
        self.close()

class UnixImageDisplay(ImageDisplay):

    """Unix socket version of image display"""

    def __init__(self, filename, family=SOCKTYPE, type=socket.SOCK_STREAM):
        ImageDisplay.__init__(self)
        try:
            self._socket = socket.socket(family, type)
            self._socket.connect(filename)
            self._fdin = self._fdout = self._socket.fileno()
        except socket.error, error:
            raise IOError("Cannot open image display")

    def close(self):

        """Close image display connection"""

        self._socket.close()


class InetImageDisplay(UnixImageDisplay):

    """INET socket version of image display"""

    def __init__(self, port, hostname=None):
        hostname = hostname or "localhost"
        UnixImageDisplay.__init__(self, (hostname, port), family=socket.AF_INET)


class ImageDisplayProxy(ImageDisplay):

    """Interface to IRAF-compatible image display

    This is a proxy to the actual display that allows retries
    on failures and can switch between display connections.
    """

    def __init__(self, imtdev=None):
        if imtdev:
            self.open(imtdev)
        else:
            self._display = None

    def open(self, imtdev=None):

        """Open image display connection, closing any active connection"""
        self.close()
        self._display = _open(imtdev)

    def close(self):

        """Close active image display connection"""

        if self._display:
            self._display.close()
            self._display = None


    def readCursor(self,sample=0):

        """Read image cursor value for the active image display

        Return immediately if sample is true, or wait for keystroke
        if sample is false (default).  Returns a string with
        x, y, frame, and key.  Opens image display if necessary.
        """

        if not self._display:
            self.open()
        try:
            value = self._display.readCursor(sample)
            # Null value indicates display was probably closed
            if value:
                return value
        except IOError, error:
                pass
        # This error can occur if image display was closed.
        # If a new display has been started then closing and
        # reopening the connection will fix it.  If that
        # fails then give up.
        self.open()
        return self._display.readCursor(sample)

    def setCursor(self,x,y,wcs):
        if not self._display:
            self.open()

        self._display.setCursor(x,y,wcs)

    def checkDisplay(self):
        """ Returns True if a valid connection to a display device is found,
        False if no connection could be found.
        """
        try:
            wcs = self._display.readInfo()
        except:
            return False
        return True



# Print help information
def help():
    print __doc__


_display = ImageDisplayProxy()

# create aliases for _display methods

readCursor = _display.readCursor
open = _display.open
close = _display.close
setCursor = _display.setCursor
checkDisplay = _display.checkDisplay



from __future__ import division # confidence high

pkg = "stsci.numdisplay"

setupargs = { 
    'version' :         "1.5",
    'description' :     "Package for displaying numpy arrays in DS9",
    'author' :          "Warren Hack",
    'maintainer_email': "help@stsci.edu",
    'url' :             "",
    'license' :         "http://www.stsci.edu/resources/software_hardware/pyraf/LICENSE",
    'platforms' :       ["any"],
    'package_dir' :     { 'stsci.numdisplay' : 'lib/stsci/numdisplay' },
    'data_files' :      [ ( 'stsci/numdisplay', ['lib/stsci/numdisplay/imtoolrc', 'LICENSE.txt','lib/stsci/numdisplay/ichar.dat'] )],
    }






"""
Version 1.0alpha - 9-Oct-2003 (WJH)

loadImtoolrc (imtoolrc=None):
    Locates, then reads in IMTOOLRC configuration file from
    system or user-specified location, and returns the 
    dictionary for reference. 

    The table gets loaded into a dictionary of the form:
        {configno:{'nframes':n,'width':nx,'height':ny},...}
    It can then be accessed using the syntax:
        fbtab[configno][attribute]
    For example:
        fbtab = loadImtoolrc()
        print fbtab[34]['width']
        1056 1024

"""    
from __future__ import division # confidence high

import os,string,sys

_default_imtoolrc_env = ["imtoolrc","IMTOOLRC"]
_default_system_imtoolrc = "/usr/local/lib/imtoolrc"
_default_local_imtoolrc = "imtoolrc"

def loadImtoolrc(imtoolrc=None):
    """
        Locates, then reads in IMTOOLRC configuration file from
        system or user-specified location, and returns the 
        dictionary for reference. 
        
    """    
    # Find the IMTOOLRC file.  Except as noted below, this order
    # matches what ximtool and ds9 use.
    _home = os.getenv("HOME")

    # Look for path to directory where this module is installed
    # This will be last-resort location for IMTOOLRC that was
    # distributed with this module.
    _module_path = os.path.split(__file__)[0]

    ####
    # list of file names to look for; ok to have None to skip an entry
    _name_list = []

    # There are two environment variables that might set the location
    # of imtoolrc:

    # getenv('imtoolrc')
    _name_list.append(os.getenv(_default_imtoolrc_env[0]))

    # getenv('IMTOOLRC')
    _name_list.append(os.getenv(_default_imtoolrc_env[1]))

    # ~/.imtoolrc
    if 'HOME' in os.environ :
        _name_list.append( os.path.join(os.environ['HOME'], ".imtoolrc") )
    _name_list.append(sys.prefix+os.sep+_default_local_imtoolrc)

    # /usr/local/lib/imtoolrc
    _name_list.append(_default_system_imtoolrc)
    
    # $iraf/dev/imtoolrc - this is not in ds9 or NOAO's ximtool,
    # but it is in the AURA Unified Release ximtool.  This is the
    # one place on your system where you can be certain that 
    # imtoolrc is really there.  Eventually, we will make a patch
    # to add this to ds9 and to IRAF.
    if 'iraf' in os.environ :
        _name_list.append( os.path.join( os.environ['iraf'], 'dev', 'imtoolrc') )

    # special to numdisplay: use imtoolrc that is in the package directory.
    # Basically, this is our way of having a built-in default table.
    _name_list.append(_module_path+os.sep+'imtoolrc')

    ####
    # Search all possible IMTOOLRC names in list
    # and open the first one found...
    for name in _name_list:
        try:
            if name:
                _fdin = open(name,'r')
                break
        except IOError, error:
            pass
    
    #Parse the file, line by line and populate the dictionary
    _lines = _fdin.readlines()
    _fdin.close()

    # Build a dictionary for the entire IMTOOL table
    # It will be indexed by configno.
    fbdict = {}
    
    for line in _lines:
        # Strip out any blanks/tabs
        line = string.strip(line)
        # Ignore empty lines
        if len(line) > 1:
            _lsp = line.split()
            # Also, ignore comment lines starting with '#'
            if _lsp[0] != '#':
                configno = int(_lsp[0])
                _dict = {'nframes':int(_lsp[1]),'width':int(_lsp[2]),'height':int(_lsp[3]),'name':_lsp[5]}
                fbdict[configno] = _dict
    return fbdict                    

def help():
    print __doc__



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distribute_setup import use_setuptools
    use_setuptools()
    from setuptools import setup

try:
    from stsci.distutils.command.easier_install import easier_install
except ImportError:
    import os
    import sys
    stsci_distutils = os.path.abspath(os.path.join('..', 'distutils', 'lib'))
    if os.path.exists(stsci_distutils) and stsci_distutils not in sys.path:
        sys.path.append(stsci_distutils)
    try:
        from stsci.distutils.command.easier_install import easier_install
        import setuptools.command.easy_install
    except ImportError:
        # If even this failed, we're not in an stsci_python source checkout,
        # so there's nothing gained from using easier_install
        from setuptools.command.easy_install import easy_install
        easier_install = easy_install
# This is required so that easier_install can be used for setup_requires
import setuptools
setuptools.command.easy_install.easy_install = easier_install

setup(
    setup_requires=['d2to1>=0.2.9', 'stsci.distutils>=0.3.2'],
    namespace_packages=['stsci'], packages=['stsci'],
    d2to1=True,
    use_2to3=True,
    zip_safe=False
)
